﻿using AIOWDMNet;
using System;
using System.Collections.Generic;
using System.Windows.Forms;
using System.Text;
using System.Threading;

namespace DIOSampleAIOWDM
{
    public partial class InputSampleForm : Form
    {
        private const int GlobalPwmDivider = 255;
        private int cardNum;
        private static readonly byte[] byteIndexToRegisterOffset = new byte[] { 0, 1, 2 };
        /// <summary>
        /// Write one byte of Digital Bit data to the specified DIO port
        /// </summary>
        /// <param name="cardNum">index of the AIOWDM-detected PCI/PCIe card to write the byte to</param>
        /// <param name="byteIndex"></param>
        /// <param name="value"></param>
        private void DIO_Write8(int cardNum, uint byteIndex, byte value) { AIOWDM.RelOutPortB( cardNum, byteIndexToRegisterOffset[byteIndex ], value); }
        private byte TTL_Read8(int cardNum, uint byteIndex) { return (byte) AIOWDM.RelInPortB( cardNum, byteIndexToRegisterOffset[byteIndex]); }
        private byte DIO_Write1(int cardNum, uint bitIndex, bool value)
        {
            /* read byte value from port */
            byte data = TTL_Read8(cardNum, bitIndex / 8);  // TODO: check for 0xAA55 and error if found

            /* perform binary arithmetic to merge new bit appropriately */
            byte mask = (byte)((uint)1 << (int)(bitIndex % 8));
            if (value == true)
                data |= mask;
            else
                data &= (byte)~mask;

            /* finish the read-modify-write operation by writing the new data value to the byte of outputs */
            DIO_Write8(cardNum, bitIndex / 8 , data);
            return 0;
        }

        public struct Info {
            public Info(string model) { Model = model; }
            public readonly string Model;
        }

        private Dictionary<uint, Info> validDeviceIDs = new Dictionary<uint, Info>();
        public InputSampleForm()
        {
            InitializeComponent();
            validDeviceIDs.Add( 0x0100, new Info("mPCIe-DIO-24S") ); // newer versions
            validDeviceIDs.Add( 0x0E57, new Info("mPCIe-DIO-24S") );
            validDeviceIDs.Add( 0x2E50, new Info("mPCIe-DIO-24A") );
        }

        private Dictionary<uint, Info> foundDeviceData = new Dictionary<uint, Info>();
        private int pollCount;
        private byte IOMask = 0b10011011;  // default to "Mode Set" and all inputs

        private void InputSampleForm_Load(object sender, EventArgs e)
        {
            var numCards = AIOWDM.GetNumCards();
            log.Items.Add("AIOWDM.GetNumCards reports " + numCards + " cards installed");

            if (numCards == 0) {log.Items.Add("Please confirm your device is installed and appears without errors in Device Manager."); } 

            for (uint numCard = 0; numCard < numCards; numCard++)
            {
                var Status = AIOWDM.QueryCardInfo((int)numCard, out uint deviceID, out uint baseAddress, out string cardName);
                if (Status == 0)
                {
                    log.Items.Add($"AIOWDM.QueryCardInfo returned FALSE (Error) for Card Number {numCard} [0x{deviceID:X4}]");
                }

                {
                    if (!validDeviceIDs.ContainsKey(deviceID))
                    {
                        cardName = cardName.TrimEnd('\0');
                        cardList.Items.Add($"CardNum #{numCard:d2}: {cardName} — NOT for this sample");
                    }
                    else
                    {
                        cardList.SelectedIndex = cardList.Items.Add($"CardNum #{numCard:d2}: {cardName}");
                        foundDeviceData.Add(numCard, validDeviceIDs[deviceID]);
                    }
                }
            }
        }

        private void CardList_SelectedIndexChanged(object sender, EventArgs e)
        {
            cardNum = cardList.SelectedIndex;
            DigitalInputPollTimer.Enabled = true;
            UpdateIOCheckboxes();
            ReadbackOutputs();
        }

        private void UpdateIOCheckboxes()
        {
            IOMask = (byte)AIOWDM.RelInPortB((int)cardNum, 3);
            cbIOG0Outputs.Checked = (IOMask & IOGroupIndexToControlBit[0]) == 0; // IOMask control bits cleared to zero indicate output mode for that I/O Group
            cbIOG1Outputs.Checked = (IOMask & IOGroupIndexToControlBit[1]) == 0;
            cbIOG2Outputs.Checked = (IOMask & IOGroupIndexToControlBit[2]) == 0;
            cbIOG3Outputs.Checked = (IOMask & IOGroupIndexToControlBit[3]) == 0;
        }

        private void ReadbackOutputs()
        {
            byte data;
            data = TTL_Read8(cardNum, 0);
            bit0.Checked = (data & 0x01) != 0;
            bit1.Checked = (data & 0x02) != 0;
            bit2.Checked = (data & 0x04) != 0;
            bit3.Checked = (data & 0x08) != 0;
            bit4.Checked = (data & 0x10) != 0;
            bit5.Checked = (data & 0x20) != 0;
            bit6.Checked = (data & 0x40) != 0;
            bit7.Checked = (data & 0x80) != 0;

            data = TTL_Read8(cardNum, 1);
            bit8.Checked = (data & 0x01) != 0;
            bit9.Checked = (data & 0x02) != 0;
            bit10.Checked = (data & 0x04) != 0;
            bit11.Checked = (data & 0x08) != 0;
            bit12.Checked = (data & 0x10) != 0;
            bit13.Checked = (data & 0x20) != 0;
            bit14.Checked = (data & 0x40) != 0;
            bit15.Checked = (data & 0x80) != 0;

            data = TTL_Read8(cardNum, 2);
            bit16.Checked = (data & 0x01) != 0;
            bit17.Checked = (data & 0x02) != 0;
            bit18.Checked = (data & 0x04) != 0;
            bit19.Checked = (data & 0x08) != 0;
            bit20.Checked = (data & 0x10) != 0;
            bit21.Checked = (data & 0x20) != 0;
            bit22.Checked = (data & 0x40) != 0;
            bit23.Checked = (data & 0x80) != 0;
        }

        private void DigitalInputPollTimer_Tick(object sender, EventArgs e)
        {
            statusEdit.Text = "poll # " + ++pollCount;
            DigitalInputPollTimer.Enabled = false;
            ReadData0.Text = Convert.ToString(TTL_Read8(cardNum, 0) ,2).PadLeft(8, '0') + " Port A read";
            ReadData1.Text = Convert.ToString(TTL_Read8(cardNum, 1), 2).PadLeft(8, '0') + " Port B read";
            ReadData2.Text = Convert.ToString(TTL_Read8(cardNum, 2), 2).PadLeft(8, '0') + " Port C read";
        }

        private void Outputs_Click(object sender, EventArgs e)
        {
        }

        private static readonly byte[] IOGroupIndexToControlBit = new byte [] { 0b00010000, 0b00000010, 0b00000001, 0b00001000 };

        private void CbIO_CheckedChanged(object sender, EventArgs e)
        {
            byte IOGroupIndex = Convert.ToByte((sender as CheckBox).Tag);
            byte controlBit = IOGroupIndexToControlBit[IOGroupIndex];
            Boolean asOutput = (sender as CheckBox).Checked;
            Boolean wasOutput = (IOMask & controlBit) == 0;

            if (wasOutput != asOutput) // to avoid spurious writes during init, only update if actually changing
            {
                if (asOutput)
                    IOMask &= (byte) ~ controlBit;   // 0 for an I/O Group's control bit configures that I/O Group for output
                else
                    IOMask |= controlBit;            // 1 sets input

                AIOWDM.RelOutPortB((int)cardNum, 0x03, IOMask);
            }

            switch (IOGroupIndex) {
                case 0:
                    OutPanelA.Enabled = asOutput; break;
                case 1:
                    OutPanelB.Enabled = asOutput; break;
                case 2:
                    OutPanelCLo.Enabled = asOutput; break;
                case 3:
                    OutPanelCHi.Enabled = asOutput; break;
            }
            ReadbackOutputs();
        }

        private void ToggleOutputBit(object sender, EventArgs e)
        {
            DIO_Write1(cardNum, Convert.ToUInt16((sender as CheckBox).Tag), (sender as CheckBox).Checked);
        }

        private void label1_Click(object sender, EventArgs e)
        {

        }

        private void textBox1_TextChanged(object sender, EventArgs e)
        {

        }

        void ConfigurePWMOutputs(int slowHz, int dutyOn, int fastHz)
        {            
            AIOWDM.RelOutPortL(0, 0x2E, GlobalPwmDivider);        // PWM divider for low freq (0->256);
            double tickRate = 125000000 / GlobalPwmDivider; // new, slower tick rate

            Console.WriteLine($"Tick Rate after 256 divider: {tickRate:f11}.  slowHz:{slowHz}, duty cycle: {dutyOn}, fastHz:{fastHz}");
            // PWM counter calculations

            int nTotalSlow = (int)tickRate/ slowHz; // how many ticks per period of slowHz
            Console.WriteLine($"total slow ticks: {nTotalSlow}; {tickRate:f11} / {slowHz}");
            int nHighSlow = (int)(dutyOn / 100.0f * nTotalSlow);
            int nLowSlow = nTotalSlow - nHighSlow;

            int nTotalFast = (int) tickRate / fastHz;
            Console.WriteLine($"total fast ticks: {nTotalFast}; {tickRate:f11} / {fastHz}");
            int nHighFast = (int)((dutyOn / 100.0f) * nTotalFast);
            int nLowFast = nTotalFast - nHighFast;

            int pulsesPerSlowHighside = (int)(nHighSlow / nTotalFast);
            Console.WriteLine($"fast Pulses per high-side of slow PWM: {pulsesPerSlowHighside}");

            
            // Configure PWM timing
            AIOWDM.RelOutPortL(0, 0x924, (uint)nHighSlow);         // Set active pulse counts on DIO8
            AIOWDM.RelOutPortL(0, 0x920, (uint)nLowSlow);          // Set inactive pulse counts
            AIOWDM.RelOutPortL(0, 0xA24, (uint)nHighFast);         // Set active pulse counts on DIO9
            AIOWDM.RelOutPortL(0, 0xA20, (uint)nLowFast);          // Set inactive pulse counts
            AIOWDM.RelOutPortL(cardNum, 0xA08, (uint)pulsesPerSlowHighside); // configure the number of Pulses in the pulse-train
            AIOWDM.RelOutPortL(0, 0xA10, 0b0000_0100);    // Set PulseHigh for high-going on non-PWM pulse-train            
        }

        void StartBothPWM()
        {
            AIOWDM.RelOutPortL(cardNum, 0x50, 0b11_0000_0000);
        }

        void StartFastPulseTrain(int freq, int dutyOn)
        {
            int nTotal = 125000000 / GlobalPwmDivider / freq;
            int nHigh = (int)(dutyOn / 100.0f * nTotal);
            int nLow = nTotal - nHigh;

            // Configure PWM timing
            AIOWDM.RelOutPortL(0, 0xA24, (uint)nHigh);         // Set active pulse counts on DIO9
            AIOWDM.RelOutPortL(0, 0xA20, (uint)nLow);          // Set inactive pulse counts
            AIOWDM.RelOutPortL(0, 0xA08, 2);
            AIOWDM.RelOutPortL(0, 0xA10, 0b0000_0100);    // Set PulseHigh for high-going on Pulse Train

            // Start PWM output on DIO9 
            AIOWDM.RelOutPortL(0, 0x50, AIOWDM.RelInPortL(0, 0x50) | 0b10_0000_0000);
        }

        void StartSlowPWM(int freq, int dutyOn)
        {
            AIOWDM.RelOutPortL(0, 0x2E, GlobalPwmDivider);        // PWM divider for low freq (0 -> 256);
            // PWM counter calculations
            int nTotal = 125000000 / GlobalPwmDivider / freq;
            int nHigh = (int)(dutyOn / 100.0f * nTotal);
            int nLow = nTotal - nHigh;
            Console.WriteLine($"{freq}Hz is {nTotal} ticks; nHigh:{nHigh}, nLow:{nLow}");

            // Configure PWM timing
            AIOWDM.RelOutPortL(0, 0x924, (uint)nHigh);         // Set active pulse counts on DIO8
            AIOWDM.RelOutPortL(0, 0x920, (uint)nLow);          // Set inactive pulse counts   
            AIOWDM.RelOutPortL(0, 0x910, 0b00000110);    // Set PulseHigh for high-going on PWM waveform

            // Start PWM output on DIO8 
            AIOWDM.RelOutPortL(0, 0x50, AIOWDM.RelInPortL(0, 0x50) | 0b01_0000_0000);
        }

        void StartFastPWM(int freq, int dutyOn)
        {
            AIOWDM.RelOutPortL(0, 0x2E, GlobalPwmDivider);        // PWM divider for low freq (0 -> 256);
            // PWM counter calculations
            int nTotal = 125000000 / GlobalPwmDivider / freq;
            int nHigh = (int)(dutyOn / 100.0f * nTotal);
            int nLow = nTotal - nHigh;
            Console.WriteLine($"{freq}Hz is {nTotal} ticks; nHigh:{nHigh}, nLow:{nLow}");

            // Configure PWM timing
            AIOWDM.RelOutPortL(0, 0xA24, (uint)nHigh);         // Set active pulse counts on DIO9
            AIOWDM.RelOutPortL(0, 0xA20, (uint)nLow);          // Set inactive pulse counts
            AIOWDM.RelOutPortL(0, 0xA10, 0b0000_0110);    // Set PulseHigh for high-going on PWM waveform
  
            // Start PWM output on DIO9 
            AIOWDM.RelOutPortL(0, 0x50, AIOWDM.RelInPortL(0, 0x50) | 0b10_0000_0000);
        }

        void StopSlowPWM()
        {
            AIOWDM.RelOutPortL(0, 0x924, 0);   // Set active pulse counts on DIO 8
            AIOWDM.RelOutPortL(0, 0x920, 100);   // Set inactive pulse counts on DIO 8
            AIOWDM.RelOutPortL(0, 0x50, AIOWDM.RelInPortL(0, 0x50) | 0b01_0000_0000);   // Set go for DIO 8 (PB0) 
        }

        void StopFastPWM()
        {
            AIOWDM.RelOutPortL(0, 0xA24, 0);   // Set active pulse counts on DIO 9 
            AIOWDM.RelOutPortL(0, 0xA20, 100);   // Set inactive pulse counts on DIO 9
            AIOWDM.RelOutPortL(0, 0x50, AIOWDM.RelInPortL(0, 0x50) | 0b10_0000_0000);   // Set go for DIO 8 (PB0) 
        }

        void StopPWM()
        {
            StopSlowPWM();
            StopFastPWM();
        }

        bool done = false;
        private void button1_Click(object sender, EventArgs e)
        {
            AIOWDM.RelOutPortB(0, 0xFC, 0b100); // perform full card reset                                                
            AIOWDM.RelOutPortB(0, 0x3, 0b10001001); // Configure A & B as output, C as input

            StartSlowPWM(50, 25);// Start slow PWM on pin 8, 10Hz, 25% duty cycle

            Thread backThread = new Thread(new ThreadStart(() =>
            {
                byte obits = 0;
                
                Thread.CurrentThread.IsBackground = true;
                Console.WriteLine(" -- Background Thread Starting");
                do
                {
                    {
                        byte bits = (byte)AIOWDM.RelInPortB(cardNum, 2); // read port C
                        if ((bits&1) != (obits&1)){
                            obits = bits;
                            switch (bits & 1)
                            {
                                case 0:
                                    StopFastPWM();
                                    break;
                                case 1:
                                    StartFastPWM(500, 50);
                                    break;
                                default:
                                    break;
                            }
                        }
                    }
                } while (!done);
                Console.WriteLine(" -- Background Thread ending");
            }));
            backThread.Priority = ThreadPriority.Highest;
            backThread.Start();

            Console.WriteLine($"FPGA Version reported:{AIOWDM.RelInPortL(0, 0x98):x8}");



            //StartFastPWM(500, 50); // Start fast PWM on pin 9

            //ConfigurePWMOutputs(50, 25, 500);
            //StartBothPWM();

            // Enable debounce
            // AIOWDM.RelOutPortB(0, 0x1100, 0b1);

            //AIOWDM.RelOutPortL(0, 0x1104, 0b0000_0011); // Generate an event when rising low-to-high or high-to-low is detected on DIO16 PC0

            // Enable IRQ
            //AIOWDM.RelOutPortB(cardNum, 0x40, 0xFF);
            //AIOWDM.RelOutPortL(0, 0x30, 1 << 16); // Enable IRQ for DIO16 (bit 16)
            //AIOWDM.RelOutPortB(0, 0x28, 0xFF); // Global IRQ enable
            //AIOWDM.RelOutPortB(cardNum, 0x0f, 0xff); // clear any pending IRQ
            //AIOWDM.RelOutPortB(cardNum, 0x0b, 0x00); // unmask all generic CoS IRQs
        }

        private void StopButton_Click(object sender, EventArgs e)
        {
            done = true;
            AIOWDM.AbortRequest(cardNum);
            StopPWM();
        }
    }
}
